#!/usr/bin/env python
"""
Tests for the `edx-rbac` models module.
"""


class TestUserRole:
    """
    Tests of the UserRole model.
    """

    def test_something(self):
        """TODO: Write real test cases."""


class TestRole:
    """
    Tests of the Role model.
    """

    def test_something(self):
        """TODO: Write real test cases."""
